DEFAULT_SERVER_HOST = 'localhost'
DEFAULT_SERVER_PORT = 11211
DEFAULT_POOL_MINSIZE = 2
DEFAULT_POOL_MAXSIZE = 5
DEFAULT_TIMEOUT = 1

STORED = b'STORED\r\n'
NOT_STORED = b'NOT_STORED\r\n'
EXISTS = b'EXISTS\r\n'
NOT_FOUND = b'NOT_FOUND\r\n'
DELETED = b'DELETED\r\n'
TOUCHED = b'TOUCHED\r\n'

END = b'END\r\n'
VERSION = b'VERSION'
OK = b'OK\r\n'
